"""
The ``utils`` module contains a set of functions which are used to perform some common tasks in the audio processing. The most important functions are:
"""

from enum import Enum
import numpy as np


class Criterion(Enum):
    """
    Represents a mathematical criterion. Some kinds of calculations can be done in different ways and often we want to have control over the way they are done. For example, when calculating the RMS of a signal, we can either take the maximum value of the RMS of each channel, or the mean value of the RMS of each channel.

    The following values are available:

    * ``max``: which value is the string ``'max'``.
    * ``min``: which value is the string ``'min'``.
    * ``mean``: which value is the string ``'mean'``.

    """

    max = 'max'
    min = 'min'
    mean = 'mean'


def db_to_pcm(db: int, bit_depth: int) -> int:
    r"""
    Convert a dB value to a float value.

    Args:
        db (int): The dB value to convert.
        bit_depth (int): The bit depth of the PCM value, this is used to calculate the maximum value of the PCM value.

    The dB value is converted to a PCM value using the following formula:

    .. math::

        \text{PCM} = \left\lfloor 2^{bit\_depth - 1} \times 10^{\frac{db}{20}}\right\rceil

    where :math:`\lfloor x \rceil` is the *round* function, which approximates the result to the nearest integer (it introduces the quantization error).
    """
    return round(10**(db / 20) * 2**(bit_depth - 1))


def pcm_to_db(pcm, bit_depth: int) -> float:
    r"""
    Convert the given signal power to a dB value.

    Args:
        pcm (int): The PCM value to convert.
        bit_depth (int): The bit depth of the PCM value, this is used to calculate the maximum value of the PCM value.

    The PCM value is converted to a dB value using the following formula:

    .. math::

        db = 20 \times log_{10}\left(\frac{\text{PCM}}{2^{bit\_depth - 1}}\right)

    .. note::

        Due to the formula the result on input ``0`` should be ``-inf``, nonetheless, as we are threating discrete quantities, we return the minimum value that can be represented by the given bit depth based on the following table:

        +---------+--------------------+
        | Bit     | Minimum value (dB) |
        +=========+====================+
        | 16      | -98                |
        +---------+--------------------+
        | 24      | -146               |
        +---------+--------------------+
        | 32      | -194               |
        +---------+--------------------+

        for a more detailed explanation see about `audio bit depth <https://en.wikipedia.org/wiki/Audio_bit_depth#Quantization>`_.
    """

    MIN_VAL = {
        "16": -98,
        "24": -146,
        "32": -194,
    }

    if pcm == 0:
        return MIN_VAL[str(bit_depth)]

    return 20 * np.log10(abs(pcm / 2**(bit_depth - 1)))


def rms(array: np.ndarray) -> float:
    r"""
    Calculate the RMS of the given array.

    The Root Mean Square (RMS) is the square root of the mean of the squares of the values in the array. It is a measure of the magnitude of a signal. It is calculated using the following formula:

    .. math::

        \text{RMS} = \sqrt{\frac{1}{n} \sum_{i=1}^{n} x_i^2}

    where :math:`x_i` is the value of the array at index :math:`i` and :math:`n` is the length of the array.
    """

    mean_squares = np.square(array).mean()
    if mean_squares == 0:
        return 0
    return np.sqrt(mean_squares)


def get_last_index(haystack: np.ndarray, threshold) -> int | tuple | None:
    """
    Get the index of the last occurrence of a value greater than ``threshold`` in the given array

    Args:
        haystack (np.ndarray): the array to inspect
        threshold (int): the threshold to use

    Returns:
        The index of the last occurrence of a value greater than the given limit in the array, if the array is multi-dimensional it returns a tuple. If there is no match, None is returned
    """
    positions = np.where(haystack >= threshold)
    indexes = positions[0]
    if indexes.size == 0:
        return None
    if len(positions) > 1:
        return list(zip(*positions))[-1]
    return indexes[-1]



import wave
import os
from os import path
import numpy as np
import librosa

from mpai_cae_arp.audio._noise import Noise
from mpai_cae_arp.audio import utils


class AudioWave:
    """
    A class to represent an audio wave.

    Parameters
    ----------

    bit : int
        the bit depth of the audio
    channels : int
        the number of channels of the audio
    samplerate : int
        the sample rate of the audio
    data : numpy.ndarray
        the data of the audio as a numpy array

    Raises
    ------
    ValueError
        if ``bit_depth`` is not 8, 16, 24 or 32
    ValueError
        if ``sample_rate`` is not between 8000 and 192000

    Examples
    --------

    Here is an example of how to create a new AudioWave object containing a sine wave with a frequency of 440 Hz, 44100 Hz sample rate, 16 bit depth and 1 channel:

    .. testcode::

        import numpy as np
        from audio import AudioWave

        freq = 440   # Hz
        sr = 44100   # sample rate
        bit = 16     # bit depth
        channels = 1 # number of channels

        # create an array of 1 second length
        time = np.arange(0, 1, 1/sr)

        # create a sine wave
        signal = np.sin(2 * np.pi * freq * time)

        # set the signal to the maximum value of the bit depth
        signal *= 2 ** (bit - 1)
        signal = np.int16(signal)

        # create the AudioWave object
        audio = AudioWave(signal, bit, channels, sr)

        print(audio)

    the code above will print the following output:

    .. testoutput::

        <AudioWave(bit: 16, chn: 1, sr: 44100)>

    """

    array: np.ndarray
    bit: int
    samplerate: int
    channels: int

    def __init__(self, data: np.ndarray, bit: int, channels: int, samplerate: int):
        if bit not in [8, 16, 24, 32]:
            raise ValueError("bit must be 8, 16, 24 or 32")
        if samplerate < 8000 or samplerate > 192000:
            raise ValueError("samplerate must be between 8 and 192 kHz")
        self.bit = bit
        self.channels = channels
        self.samplerate = samplerate
        self.array = data

    def __len__(self):
        return len(self.array)

    def __iter__(self):
        return iter(self.array)

    def __getitem__(self, key):
        return AudioWave(self.array[key], self.bit, self.channels, self.samplerate)

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, AudioWave):
            return np.array_equal(
                self.array, __o.array
            ) and self.bit == __o.bit and self.channels == __o.channels and self.samplerate == __o.samplerate
        raise TypeError(f"cannot compare AudioWave with {type(__o)}")

    def __repr__(self):
        return f"<AudioWave(bit: {self.bit}, chn: {self.channels}, sr: {self.samplerate})>"

    @staticmethod
    def from_file(filepath: str,
                  bit: int = None,
                  channels: int = None,
                  samplerate: int = None,
                  bufferize: bool = False):
        """
        A static method to generate an AudioWave object from a file. It reads the binary headers of the file to automatically get the bit depth, the number of channels and the sample rate. If any of these values is given, it will be used instead of the one read from the file.
        """

        if bufferize:
            data = None
            for chunk in AudioWave.buffer_generator_from_file(filepath):
                if data is None:
                    bit = chunk.bit
                    channels = chunk.channels
                    samplerate = chunk.samplerate
                    data = chunk.array
                else:
                    data = np.concatenate((data, chunk.array), axis=0)

            return AudioWave(data, bit, channels, samplerate)

        with wave.open(filepath, 'rb') as fp:
            raw_data = fp.readframes(fp.getnframes())
            samplerate = fp.getframerate() if samplerate is None else samplerate
            channels = fp.getnchannels() if channels is None else channels
            bit = fp.getsampwidth() * 8 if bit is None else bit

        return AudioWave.from_bytes(raw_data, bit, channels, samplerate)

    @staticmethod
    def read_file_metadata(filepath: str):
        """
        A static method to read the metadata of an audio file. It reads the binary headers of the file to automatically get the bit depth, the number of channels and the sample rate.
        """
        with wave.open(filepath, 'rb') as fp:
            samplerate = fp.getframerate()
            channels = fp.getnchannels()
            bit = fp.getsampwidth() * 8

        return bit, channels, samplerate

    @staticmethod
    def buffer_generator_from_file(filepath: str, buffer_size: int = 1024 * 1024 * 8):
        """Return a generator that yields AudioWave objects from a file. The generator will read the file in chunks of `buffer_size` bytes.

        Parameters
        ----------
        filepath: str
            The path to the file.
        buffer_size: int
            The size of audio chunks. Defaults to 1024*1024*8 (8 MegaBytes).

        Yields
        ------
        AudioWave
            An AudioWave object containing the audio data read from the file.
        """
        bit_dept, chn, sr = AudioWave.read_file_metadata(filepath)
        with wave.open(filepath, 'rb') as wave_file:

            frames: bytes = wave_file.readframes(buffer_size)
            while frames != b'':
                yield AudioWave.from_bytes(frames, bit_dept, chn, sr)
                frames = wave_file.readframes(buffer_size)

    @staticmethod
    def from_bytes(raw_data: bytes, bit: int, channels: int, samplerate: int):
        """A static method to generate an AudioWave object from a stream of bytes. It is assumed that the data is in little endian format and that the bytes are signed integers.

        Parameters
        ----------
        raw_data: bytes
            The stream of bytes.
        bit: int
            The bit depth of the audio.
        channels: int
            The number of channels of the audio.
        samplerate: int
            The sample rate of the audio.

        Example
        -------

        The following example shows how to create an AudioWave object from a stream of bytes.

        .. doctest::

            >>> import numpy as np
            >>> from audiohandler import AudioWave
            >>> audio = AudioWave.from_bytes(b'\\x00\\x00\\x00\\x00', 16, 2, 44100)
            >>> np.array_equal(audio.array, np.array([[0, 0]]))
            True

        """
        data = np.array([
            int.from_bytes(raw_data[i:i + bit // 8], byteorder='little', signed=True)
            for i in range(0, len(raw_data), bit // 8)
        ], dtype=np.int64)  # int32 (default in Windows) leds to overflow
        data = np.reshape(data, (-1, channels))
        return AudioWave(data, bit, channels, samplerate)

    def save(self, filepath: str, force: bool = False):
        """Save the audio as a wave file at the given path.

        Parameters
        ----------
        filepath: str
            The path where to save the audio.
        force: bool
            If True, it will create the directory if it does not exist.

        Raises
        ------
        ValueError
            if the directory does not exist and force is False

        Examples
        --------

        .. code-block:: python

            from audio import AudioWave

            audio = AudioWave(np.array([[0, 0]]), 16, 2, 44100)
            audio.save('wrong/path/test.wav') # raises ValueError
            audio.save('force/path/test.wav', force=True) # creates the path and saves the file

        """

        if not path.exists(path.dirname(filepath)):
            if force:
                os.makedirs(path.dirname(filepath))
            else:
                raise ValueError(f"Directory {path.dirname(filepath)} does not exist")

        with wave.open(filepath, 'wb') as fp:
            fp.setframerate(self.samplerate)
            fp.setnchannels(self.channels)
            fp.setsampwidth(self.bit // 8)
            fp.setnframes(self.number_of_frames())
            fp.writeframesraw(self.get_raw())

    def set_sample_rate(self, samplerate: int):
        """Set the sample rate of the audio.

        .. versionadded:: 0.4.0

        Parameters
        ----------
        samplerate: int
            The new sample rate.
        """
        self.array = librosa.resample(self.array, self.samplerate, samplerate)
        self.samplerate = samplerate

    def get_raw(self) -> bytes:
        """Get the raw data of the audio.

        Returns:
            A bytes stream in the form (left right left right ...), where left and right are the samples for the left and right channels respectively for each frame in little endian format and signed.
        """
        data: np.ndarray = np.reshape(self.array, (-1, ))
        bytes_array = b''.join([
            int(x).to_bytes(self.bit // 8, byteorder='little', signed=True)
            for x in data
        ])

        return bytes_array

    def number_of_frames(self) -> int:
        """Get the number of frames in the audio. A frame is a sample for each channel.
        In a stereo audio, a frame is composed by two samples,
        one for each channel, in the form [left, right],
        where left and right are the samples for the left and right channels respectively.

        Returns:
            The number of frames in the audio
        """
        return len(self.array)

    def duration(self) -> float:
        r"""Get the duration of the audio in seconds.
        The duration is calculated by dividing the number of frames by the sample rate.

        .. math::

            time = \frac{frames}{sample\_rate}

        Returns
        -------
        float
            the duration of the audio in seconds
        """
        return self.number_of_frames() / self.samplerate

    def get_mfcc(self, n_mfcc: int = 13) -> np.ndarray:
        """Get the Mel-frequency cepstral coefficients (MFCC) of the audio.

        Parameters
        ----------
        n_mfcc: int
            The number of MFCC to return. Default is 13.

        Returns
        -------
        np.ndarray
            The mean of the first ``n_mfcc`` mfcc over the entire audio signal. If the audio is multichannel, the mean of the mfcc of each channel is returned.
        """

        mfcc_per_ch = []

        for channel in range(self.channels):
            signal = self.get_channel(channel).array
            signal = signal / (2 ^ (self.bit - 1))  # normalize the signal
            mfccs = librosa.feature.mfcc(y=signal, sr=self.samplerate, n_mfcc=n_mfcc)
            mean_mfccs = []
            for e in mfccs:
                mean_mfccs.append(np.mean(e))
            mfcc_per_ch.append(mean_mfccs)

        # return the mean of the mfcc of each channel
        return np.mean(mfcc_per_ch, axis=0)

    def rms(self, mode: utils.Criterion = utils.Criterion.max) -> float:
        r"""The root mean square of the audio. It measures the power level of the entire signal. It is calculated by taking the square root of the mean of the square of the samples. In multichannel audio, the rms is calculated by taking the mean or the max of the rms of each channel.

        Parameters
        ----------
        mode: Criterion
            The mode to use to calculate the rms. It can be ``Criterion.mean`` or ``Criterion.max``. If ``Criterion.mean``, the rms is calculated by taking the mean of the rms of each channel. If ``Criterion.max``, the rms is calculated by taking the maximum of the rms between all channels. Default is ``Criterion.max``.

        Raises
        ------
        ValueError
            if the mode is not ``mean`` or ``max``

        Returns
        -------
        float
            The rms of the audio.
        """
        if self.channels == 1:
            return utils.rms(self.array)
        channels_rms = []
        for channel in range(self.channels):
            channels_rms.append(utils.rms(self.array[:, channel]))
        if mode == utils.Criterion.mean:
            return np.array(channels_rms).mean()
        if mode == utils.Criterion.max:
            return max(channels_rms)
        raise ValueError("Invalid criterion")

    def db_rms(self, mode: utils.Criterion = utils.Criterion.max) -> float:
        """
        A wrapper method for ``self.rms()`` that converts the returned value to decibels.

        .. seealso::
            see also the method :meth:`AudioWave.rms` for a more detailed description of the parameters.
        """
        return utils.pcm_to_db(self.rms(mode), self.bit)

    def get_channel(self, channel: int):
        """Create a new AudioWave object containing only the given channel.

        Parameters
        ----------
        channel: int
            the channel to extract from the audio, where 0 is the first channel, 1 the second and so on.

        Raises
        ------
        IndexError
            if the channel is not found.

        Returns
        -------
        AudioWave
            An AudioWave object containing only the given channel.

        Example
        -------

        .. code-block:: python

            from audio import AudioWave

            # let's say 'sample.wav' is a stereo audio

            # returns an AudioWave object containing only the first channel
            audio_left = AudioWave.from_file('sample.wav').get_channel(0)

            # returns an AudioWave object containing only the second channel
            audio_right = AudioWave.from_file('sample.wav').get_channel(1)

            # raises IndexError because there are only two channels
            audio_third = AudioWave.from_file('sample.wav').get_channel(2)

        """
        if channel not in range(self.channels):
            raise IndexError("Channel not found")
        return AudioWave(self.array[:, channel], self.bit, 1, self.samplerate)

    def get_silence_slices(self, noise_list: list[Noise],
                           length: int) -> dict[str, list[tuple[int, int]]]:
        """Get the slices of silence in the given signal.

        Passing a list of ``Noise`` instances to the ``noise_list`` parameter, the function will return a dictionary of slices of silence in the given signal. The dictionary will have the label of the noise as key and a list of tuples as value. Each tuple will contain the starting and ending frames of a slice of silence.

        Parameters
        ----------
        noise_list: list[Noise]
            the noise bands to use to detect the silence
        length: int
            the length of a slice of silence in milliseconds

        Raises
        ------
        ValueError
            if ``length`` is less than 1

        Returns
        -------
        dict[str, list[tuple[int, int]]]
            A dictionary with the label of the noise as key and a list of tuples as value. Each tuple will contain the starting and ending frames of a slice of silence. If the signal is completely silent, or the ``min_length`` required is greater than the signal length, the dictionary will contain an empty list for each noise in the ``noise_list``.

        Example
        -------

        For example, if we have the following audio:

        .. doctest::

            >>> from audio import AudioWave, Noise
            >>> noise_list = [Noise("noise1", -10, -20), Noise("noise2", -30, -40)]
            >>> array = np.array([10000 for _ in range(8000)])
            >>> audio = AudioWave(array, 16, 1, 8000)
            >>> audio.get_silence_slices(noise_list, 500)
            {'noise1': [(0, 4000), (4000, 8000)], 'noise2': []}


        .. note::
            The ``AudioWave`` abstraction makes possible to analyze the signal at a frame level. Nonetheless, the function will scan the signal at a millisecond interval.
            This is because in a millisecond there are many frames that can have a rms value which belongs to different ``Noise`` instances, but those variations are not relevant for the purpose of detecting silence since are not perceptible by the human ear.
        """
        # sanity checks on the parameters
        if not isinstance(length, int) or length < 1:
            raise ValueError("min_length must be an integer greater or equal 1")

        window_frames = length * self.samplerate // 1000
        last_frame = self.number_of_frames() - window_frames

        # create an empty array of indexes for each noise to filter
        idxs = {noise.label: [] for noise in noise_list}

        # if the signal is too short, return an empty dict
        if last_frame < 1 or noise_list == []:
            return idxs

        # find the thresholds to be used to detect the noise
        upper_noise_limit = max(noise.db_max for noise in noise_list)
        lower_noise_limit = min(noise.db_min for noise in noise_list)

        i = 0

        while i <= last_frame:
            chunk = self[i:i + window_frames]
            max_pos = utils.get_last_index(
                chunk.array, utils.db_to_pcm(upper_noise_limit, chunk.bit))
            # if there is a value over the threshold go to its index and start seeking from here
            if max_pos is not None:
                # if multi-dimensional array get only the x value
                if isinstance(max_pos, tuple):
                    max_pos = max_pos[0]
                i += max_pos + 1
            # else all the signal is under the threshold
            else:
                # classification of the chunk based on rms
                for noise in noise_list:
                    if noise.db_min <= chunk.db_rms() < noise.db_max:
                        idxs[noise.label].append((i, i + window_frames))
                        break
                # if the signal power is under the minimum, assign it to the quieter noise
                if chunk.db_rms() < lower_noise_limit:
                    idxs[min(noise_list).label].append((i, i + window_frames))
                i += window_frames

        return idxs



# pylint: disable=too-few-public-methods
from typing import TypeVar
import uuid
from enum import Enum
from pydantic import BaseModel, Field, ConfigDict

from mpai_cae_arp.audio.standards import EqualizationStandard, SpeedStandard
from mpai_cae_arp.files import File, FileType
from mpai_cae_arp.time import time_to_seconds


class IrregularityType(str, Enum):
    BRANDS_ON_TAPE = "b"
    SPLICE = "sp"
    START_OF_TAPE = "sot"
    ENDS_OF_TAPE = "eot"
    DAMAGED_TAPE = "da"
    DIRT = "di"
    MARKS = "m"
    SHADOWS = "s"
    WOW_AND_FLUTTER = "wf"
    PLAY_PAUSE_STOP = "pps"
    SPEED = "ssv"
    EQUALIZATION = "esv"
    SPEED_AND_EQUALIZATION = "ssv"
    BACKWARD = "sb"


class Source(str, Enum):
    AUDIO = "a"
    VIDEO = "v"
    BOTH = "b"


SelfIrregularity = TypeVar("SelfIrregularity", bound="Irregularity")
SelfIrregularityProperties = TypeVar("SelfIrregularityProperties",
                                     bound="IrregularityProperties")
SelfIrregularityFile = TypeVar("SelfIrregularityFile", bound="IrregularityFile")


class IrregularityProperties(BaseModel):
    reading_speed: SpeedStandard = Field(serialization_alias="ReadingSpeedStandard")
    reading_equalisation: EqualizationStandard = Field(
        serialization_alias="ReadingEqualisationStandard")
    writing_speed: SpeedStandard = Field(serialization_alias="WritingSpeedStandard")
    writing_equalisation: EqualizationStandard = Field(
        serialization_alias="WritingEqualisationStandard")

    @staticmethod
    def from_json(json_property: dict) -> SelfIrregularityProperties:
        return IrregularityProperties(
            reading_speed=SpeedStandard(json_property["ReadingSpeedStandard"]),
            reading_equalisation=EqualizationStandard(
                json_property["ReadingEqualisationStandard"]),
            writing_speed=SpeedStandard(json_property["WritingSpeedStandard"]),
            writing_equalisation=EqualizationStandard(
                json_property["WritingEqualisationStandard"]))

    def to_json(self) -> dict:
        """
        .. deprecated:: 0.4.0
            Use :meth:`IrregularityProperties.json` instead.
        """
        return {
            "ReadingSpeedStandard": self.reading_speed.value,
            "ReadingEqualisationStandard": self.reading_equalisation.value,
            "WritingSpeedStandard": self.writing_speed.value,
            "WritingEqualisationStandard": self.writing_equalisation.value,
        }

    def get_irregularity_type(self) -> IrregularityType | None:
        if self.reading_equalisation != self.writing_equalisation:
            if self.reading_speed != self.writing_speed:
                return IrregularityType.SPEED_AND_EQUALIZATION
            return IrregularityType.EQUALIZATION
        if self.reading_speed != self.writing_speed:
            return IrregularityType.SPEED
        return None


class Irregularity(BaseModel):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, serialization_alias="IrregularityID")
    source: Source = Field(serialization_alias="Source")
    time_label: str = Field(serialization_alias="TimeLabel")
    type: IrregularityType | None = Field(default=None, serialization_alias="IrregularityType")
    properties: IrregularityProperties | None = Field(default=None,
                                                      serialization_alias="IrregularityProperties")
    image_URI: str | None = Field(default=None, serialization_alias="ImageURI")
    audio_block_URI: str | None = Field(default=None, serialization_alias="AudioBlockURI")

    @staticmethod
    def from_json(json_irreg: dict) -> SelfIrregularity:

        properties = None
        if json_irreg.get("IrregularityProperties") is not None:
            properties = IrregularityProperties.from_json(
                json_irreg["IrregularityProperties"])

        raw_irreg_type = json_irreg.get("IrregularityType")
        irregularity_type = None
        if raw_irreg_type is not None:
            if raw_irreg_type is not (IrregularityType.SPEED.value
                                      or IrregularityType.SPEED_AND_EQUALIZATION.value):
                irregularity_type = IrregularityType(raw_irreg_type)
            else:
                if properties.reading_equalisation != properties.writing_equalisation:
                    irregularity_type = IrregularityType.SPEED_AND_EQUALIZATION
                else:
                    irregularity_type = IrregularityType.SPEED

        return Irregularity(id=uuid.UUID(json_irreg["IrregularityID"]),
                            source=Source(json_irreg["Source"]),
                            time_label=json_irreg["TimeLabel"],
                            type=irregularity_type,
                            properties=properties,
                            image_URI=json_irreg.get("ImageURI"),
                            audio_block_URI=json_irreg.get("AudioBlockURI"))

    def to_json(self) -> dict:
        """
        Returns a dictionary with the irregularity information

        .. deprecated:: 0.4.0
            Use :func:`Irregularity.json` instead.
        """
        dictionary = {
            "IrregularityID": str(self.id),
            "Source": self.source.value,
            "TimeLabel": self.time_label,
        }

        if self.type:
            dictionary["IrregularityType"] = self.type.value

        if self.image_URI:
            dictionary["ImageURI"] = self.image_URI

        if self.audio_block_URI:
            dictionary["AudioBlockURI"] = self.audio_block_URI

        if self.properties:
            dictionary["IrregularityProperties"] = self.properties.to_json()

        return dictionary


class IrregularityFile(BaseModel):
    irregularities: list[Irregularity] = Field(serialization_alias="Irregularities")
    offset: int | None = Field(default=None, serialization_alias="Offset")

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, IrregularityFile):
            return False

        return self.irregularities == __o.irregularities and self.offset == __o.offset

    @staticmethod
    def from_json(json_irreg: dict) -> SelfIrregularityFile:
        irregularities = []

        for irreg in json_irreg["Irregularities"]:
            irregularities.append(Irregularity.from_json(irreg))

        return IrregularityFile(irregularities=irregularities,
                                offset=json_irreg.get("Offset"))

    def to_json(self) -> dict:
        dictionary = {
            "Irregularities":
                [irregularity.to_json() for irregularity in self.irregularities],
        }

        if self.offset:
            dictionary["Offset"] = self.offset

        return dictionary

    def add(self, irregularity: Irregularity) -> SelfIrregularityFile:
        """Add an irregularity to the list of irregularities.

        Parameters
        ----------
        irregularity : Irregularity
            the irregularity to add

        Raises
        ------
        TypeError
            if the irregularity is not a py:class:`Irregularity` object
        """
        if not isinstance(irregularity, Irregularity):
            raise TypeError("IrregularityFile.add() expects an Irregularity object")
        self.irregularities.append(irregularity)
        self.order()
        return self

    def order(self) -> SelfIrregularityFile:
        self.irregularities.sort(key=lambda x: time_to_seconds(x.time_label))
        return self

    def join(self, other) -> SelfIrregularityFile:
        """Append the irregularities of other in current irregularity file, ordered by time.

        Parameters
        ----------
        other : IrregularityFile
            the irregularity file you want to append at the current one

        Raises
        ------
        TypeError
            if other is not an instance of IrregularityFile
        """
        if not isinstance(other, IrregularityFile):
            raise TypeError("other must be an instance of IrregularityFile")
        self.irregularities += other.irregularities
        self.order()
        return self

    def save_as_json_file(self, path: str) -> None:
        """
        Save the irregularity file as a JSON file at the given path.

        .. versionadded:: 0.4.0
        """
        File(path=path,
             format=FileType.JSON).write_content(self.model_dump(by_alias=True))



from .irregularity import (
    Irregularity,
    IrregularityFile,
    IrregularityProperties,
    IrregularityType,
    Source,
)
from .restoration import Restoration, EditingList
from . import schema

__all__ = [
    "Irregularity",
    "IrregularityFile",
    "IrregularityProperties",
    "IrregularityType",
    "Restoration",
    "EditingList",
    "schema",
    "Source",
]



from io import TextIOWrapper
from enum import Enum
import json
import yaml
import codecs
from pydantic import BaseModel, field_validator


class FileAction(str, Enum):
    READ = "r"
    WRITE = "w"
    APPEND = "a"


class FileType(str, Enum):
    YAML = "yaml"
    JSON = "json"


class File(BaseModel):
    encoding: str = "utf-8"
    format: FileType
    path: str

    @field_validator('encoding')
    def encoding_validator(cls, v: str):
        try:
            codecs.lookup(v)
        except LookupError:
            raise ValueError(f"Encoding {v} not supported")
        return v

    def open(self, action: FileAction) -> TextIOWrapper:
        """
        Open the file with the given action        

        Parameters
        ----------
        action : FileAction
            the action to perform on the file at opening

        Returns
        -------
        TextIOWrapper
            the file descriptor
        """
        return open(self.path, action.value, encoding=self.encoding)

    def get_content(self) -> dict:
        """
        Return the file content

        Raises
        ------
        ValueError
            if the format is not supported
        
        Returns
        -------
        dict
            the parsed content of the file
        """
        with self.open(FileAction.READ) as fd:
            if self.format == FileType.YAML:
                content = yaml.safe_load(fd)
                return content
            if self.format == FileType.JSON:
                content = json.load(fd)
            else:
                raise ValueError("Format not supported")

        return content

    def write_content(self, content: dict | str) -> None:
        """
        Write the given content in the file

        Parameters
        ----------
        content : dict | str
            the content to write in the file
            if content is a string, it will be parsed as a JSON string
        """

        if isinstance(content, str):
            try:
                content = json.loads(json.loads(content))
            except json.decoder.JSONDecodeError:
                raise ValueError("Content str is not a valid JSON string")

        with self.open(FileAction.WRITE) as fd:
            if self.format == FileType.YAML:
                yaml.safe_dump(content, fd)
            elif self.format == FileType.JSON:
                json.dump(content, fd, indent=4)
            else:
                raise ValueError("Format not supported")



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import arp_pb2 as arp__pb2


class AIMStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.getInfo = channel.unary_unary(
            '/arp.AIM/getInfo',
            request_serializer=arp__pb2.InfoRequest.SerializeToString,
            response_deserializer=arp__pb2.InfoResponse.FromString,
        )
        self.work = channel.unary_stream(
            '/arp.AIM/work',
            request_serializer=arp__pb2.JobRequest.SerializeToString,
            response_deserializer=arp__pb2.JobResponse.FromString,
        )


class AIMServicer(object):
    """Missing associated documentation comment in .proto file."""

    def getInfo(self, request, context):
        """
        Get information about the AIM
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def work(self, request, context):
        """
        A generic method to perform the AIM task
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AIMServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'getInfo':
        grpc.unary_unary_rpc_method_handler(
            servicer.getInfo,
            request_deserializer=arp__pb2.InfoRequest.FromString,
            response_serializer=arp__pb2.InfoResponse.SerializeToString,
        ),
        'work':
        grpc.unary_stream_rpc_method_handler(
            servicer.work,
            request_deserializer=arp__pb2.JobRequest.FromString,
            response_serializer=arp__pb2.JobResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler('arp.AIM',
                                                           rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler, ))


# This class is part of an EXPERIMENTAL API.
class AIM(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def getInfo(request,
                target,
                options=(),
                channel_credentials=None,
                call_credentials=None,
                insecure=False,
                compression=None,
                wait_for_ready=None,
                timeout=None,
                metadata=None):
        return grpc.experimental.unary_unary(request, target, '/arp.AIM/getInfo',
                                             arp__pb2.InfoRequest.SerializeToString,
                                             arp__pb2.InfoResponse.FromString, options,
                                             channel_credentials, insecure,
                                             call_credentials, compression,
                                             wait_for_ready, timeout, metadata)

    @staticmethod
    def work(request,
             target,
             options=(),
             channel_credentials=None,
             call_credentials=None,
             insecure=False,
             compression=None,
             wait_for_ready=None,
             timeout=None,
             metadata=None):
        return grpc.experimental.unary_stream(request, target, '/arp.AIM/work',
                                              arp__pb2.JobRequest.SerializeToString,
                                              arp__pb2.JobResponse.FromString, options,
                                              channel_credentials, insecure,
                                              call_credentials, compression,
                                              wait_for_ready, timeout, metadata)



# pylint: disable=too-few-public-methods
import uuid
from typing import TypeVar
from pydantic import BaseModel, Field, ConfigDict
from mpai_cae_arp.audio.standards import SpeedStandard, EqualizationStandard
from mpai_cae_arp.files import File, FileType


class Restoration(BaseModel):
    """
    .. versionadded:: 0.4.0
    """

    id: uuid.UUID = Field(default_factory=uuid.uuid4, serialization_alias="RestorationID")
    preservation_audio_file_start: str = Field(serialization_alias="PreservationAudioFileStart")
    preservation_audio_file_end: str = Field(serialization_alias="PreservationAudioFileEnd")
    restored_audio_file_URI: str = Field(serialization_alias="RestoredAudioFileURI")
    reading_backwards: bool = Field(serialization_alias="ReadingBackwards")
    applied_speed_standard: SpeedStandard = Field(serialization_alias="AppliedSpeedStandard")
    applied_sample_frequency: int = Field(serialization_alias="AppliedSampleFrequency")
    applied_equalization_standard: EqualizationStandard = Field(
        serialization_alias="AppliedEqualisationStandard")


Self = TypeVar("Self", bound="EditingList")


class EditingList(BaseModel):
    """
    .. versionadded:: 0.4.0
    """

    original_speed_standard: SpeedStandard = Field(serialization_alias="OriginalSpeedStandard")
    original_equalization_standard: EqualizationStandard = Field(
        serialization_alias="OriginalEqualisationStandard")
    original_sample_frequency: int = Field(serialization_alias="OriginalSampleFrequency")
    restorations: list[Restoration] = Field(serialization_alias="Restorations")

    def add(self, restoration: Restoration) -> Self:
        self.restorations.append(restoration)
        return self

    def remove(self, restoration: Restoration) -> Self:
        self.restorations.remove(restoration)
        return self

    def remove_by_id(self, restoration_id: uuid.UUID) -> Self:
        filtered = list(filter(lambda r: r.id != restoration_id, self.restorations))

        if len(filtered) == len(self.restorations):
            raise ValueError(f"Restoration with ID {restoration_id} not found.")

        self.restorations = filtered
        return self

    def save_as_json_file(self, path: str) -> None:
        File(path=path,
             format=FileType.JSON).write_content(self.model_dump(mode='json', by_alias=True))



"""
This module contains functions to convert time formats. The agreed time format is hh:mm:ss.msc as a string in MPAI standards.
"""
import datetime


def seconds_to_string(seconds: float) -> str:
    """Converts seconds to a human-readable time format (hh:mm:ss.msc).

    Parameters
    ----------
    seconds : float
        The number of seconds to convert

    Returns
    -------
    str
        A human-readable time format
    """

    seconds = float(seconds)

    # convert seconds in the format hh:mm:ss.msc
    time = str(datetime.timedelta(seconds=seconds))

    if seconds.is_integer():
        time = time + ".000"

    # convert days to hours
    if "day" in time:
        split_time = time.split(", ")
        days = int(split_time[0].split(" ")[0])
        hours = int(split_time[1].split(":")[0])
        time = f'{days * 24 + hours}:{":".join(split_time[1].split(":")[1::])}'

    time = ':'.join([x.zfill(2) for x in time.split(":")])  # add leading zeros
    split_time = time.split(".")
    time = split_time[0] + "." + split_time[1][:3]  # msc precision is 3

    return time


def time_to_seconds(time: str) -> float:
    """
    Converts a time string in the format hh:mm:ss.msc in seconds.

    Parameters
    ----------
    time : str
        The time string to convert

    Raises
    ------
    ValueError
        If the time string is not in the correct format

    Returns
    -------
    float
        The number of seconds

    """

    if not time.replace(":", "").replace(".", "").isdigit():
        raise ValueError("The time string is not in the correct format")

    split_time = time.split(":")
    hours = int(split_time[0])
    minutes = int(split_time[1])
    seconds = float(split_time[2])

    return hours * 3600 + minutes * 60 + seconds


def frames_to_seconds(frames: int, fps: int) -> float:
    """
    Converts a number of frames in seconds.

    Parameters
    ----------
    frames : int
        The number of frames
    fps : int
        The number of frames per second

    Raises
    ------
    ValueError
        If the number of frames per second is less than or equal to 0 or if the number of frames is less than 0

    Returns
    -------
    float
        The number of seconds

    """

    if fps <= 0:
        raise ValueError("The number of frames per second must be greater than 0")
    if frames < 0:
        raise ValueError("The number of frames must be greater than or equal to 0")

    return frames / fps


def seconds_to_frames(seconds: float, fps: int) -> int:
    """
    Converts a number of seconds in frames.

    Parameters
    ----------
    seconds : float
        The number of seconds
    fps : int
        The number of frames per second

    Raises
    ------
    ValueError
        If the number of frames per second is less than or equal to 0 or if the number of seconds is less than 0

    Returns
    -------
    int
        The number of frames

    """

    if fps <= 0:
        raise ValueError("The number of frames per second must be greater than 0")
    if seconds < 0:
        raise ValueError("The number of seconds must be greater than or equal to 0")

    return int(seconds * fps)



__author__ = "Matteo Spanio"
__copyright__ = "Copyright 2023, Audio Innova S.r.l."
__credits__ = ["Niccolò Pretto", "Nadir Dalla Pozza", "Sergio Canazza"]
__license__ = "GPL v3.0"
__maintainer__ = "Matteo Spanio"
__email__ = "dev2@audioinnova.com"
__status__ = "Development"



"""
This module contains functions to print and format text in the console. Consider it as deprecated, use instead rich library.
"""
from enum import Enum

END = '\033[0m'


class Style(Enum):
    """
    Styles for console output
    """
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class Color(Enum):
    """
    Colors for console output
    """
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARK_CYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'


def prettify(text: str, color: Color = None, styles: list[Style] = None) -> str:
    """
    Formats a string with some styles

    Parameters
    ----------
    text : str
        string to format
    color : Color, optional
        color to use
    styles : list[Style], optional
        styles to use
    
    Returns
    -------
    str
        formatted string
    """

    prepend = ''

    if color:
        prepend += color.value
    if styles:
        for style in styles:
            prepend += style.value

    return prepend + text + END


def pprint(text: str, color: Color = None, styles: list[Style] = None) -> None:
    """
    Formats a string with some styles and prints it

    Parameters
    ----------
    text : str
        string to format
    color : Color, optional
        color to use
    styles : list[Style], optional
        styles to use
    """

    print(prettify(text, color, styles))



# pylint: disable=too-few-public-methods
from pydantic import ConfigDict, BaseModel, Field


class Contact(BaseModel):
    name: str = Field(..., description="Name of the contact person.")
    email: str = Field(..., description="Email of the contact person.")
    model_config = ConfigDict(json_schema_extra={"example": {"name": "John Doe", "email": "email@email.com"}})


class License(BaseModel):
    name: str = Field(..., description="Name of the license.")
    url: str = Field(..., description="URL of the license.")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "name": "MIT",
            "url": "https://opensource.org/licenses/MIT"
        }
    })


class Info(BaseModel):
    title: str = Field(..., description="The title of the API.")
    description: str = Field(..., description="A short description of the API.")
    version: str = Field(..., description="The version of the API.")
    contact: Contact = Field(
        ..., description="Contact information for the owners of the API.")
    license_info: License = Field(..., description="License information for the API.")
    model_config = ConfigDict(json_schema_extra={
        "example": {
            "title": "API title",
            "description": "A very nice API",
            "version": "0.1.0",
            "contact": {
                "name": "John Doe",
                "email": "example@example.com",
            },
            "license_info": {
                "name": "MIT",
                "url": "https://opensource.org/licenses/MIT",
            }
        }
    })



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: arp.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\tarp.proto\x12\x03\x61rp\"+\n\x0bInfoRequest\x12\x12\n\x05\x66ield\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x08\n\x06_field\"S\n\nJobRequest\x12\x13\n\x0bworking_dir\x18\x01 \x01(\t\x12\x12\n\nfiles_name\x18\x02 \x01(\t\x12\x12\n\x05index\x18\x03 \x01(\x05H\x00\x88\x01\x01\x42\x08\n\x06_index\".\n\x0bJobResponse\x12\x0e\n\x06status\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\"&\n\x07\x43ontact\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x65mail\x18\x02 \x01(\t\"$\n\x07License\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\"\x81\x01\n\x0cInfoResponse\x12\r\n\x05title\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\x12\x1d\n\x07\x63ontact\x18\x04 \x01(\x0b\x32\x0c.arp.Contact\x12\x1d\n\x07license\x18\x05 \x01(\x0b\x32\x0c.arp.License2f\n\x03\x41IM\x12\x30\n\x07getInfo\x12\x10.arp.InfoRequest\x1a\x11.arp.InfoResponse\"\x00\x12-\n\x04work\x12\x0f.arp.JobRequest\x1a\x10.arp.JobResponse\"\x00\x30\x01\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'arp_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _INFOREQUEST._serialized_start = 18
    _INFOREQUEST._serialized_end = 61
    _JOBREQUEST._serialized_start = 63
    _JOBREQUEST._serialized_end = 146
    _JOBRESPONSE._serialized_start = 148
    _JOBRESPONSE._serialized_end = 194
    _CONTACT._serialized_start = 196
    _CONTACT._serialized_end = 234
    _LICENSE._serialized_start = 236
    _LICENSE._serialized_end = 272
    _INFORESPONSE._serialized_start = 275
    _INFORESPONSE._serialized_end = 404
    _AIM._serialized_start = 406
    _AIM._serialized_end = 508
# @@protoc_insertion_point(module_scope)



from enum import Enum


class EqualizationStandard(str, Enum):
    IEC = "IEC"
    CCIR = "IEC1"
    NAB = "IEC2"


class SpeedStandard(float, Enum):
    I = 0.9375
    II = 1.875
    III = 3.75
    IV = 7.5
    V = 15
    VI = 30
